/*  vmath.c:   variable length infinite precision integer arithmetic */

#include "eclib.h"

/*
 * vadd() variable length integer addition
 * vsub() variable length integer subtraction
 * vmul() variable length integer multiplication
 * vdiv() variable length integer division
 *
 * If you alter these routines, or convert them to assembly for efficiency,
 * a good regression test is the command
 * factor 9938377282939377229933889992299
 * in the pbc program.
 * The assembly language file provided (MSDOS, Microsoft C/assembly, small memory
 * model) runs 4 times as fast as these routines.
 * It is of course 4 times as difficult to debug and maintain.
 *
 *  Each number is sign magnitude, where the sign is determined by the high
 *  order bit of VINT.size.  See struct VINT in eclib.h.
 *  Thus all numbers are limited in absolute value to 32767*16 bits.
 *  
 *  We always assume there is enough space for the result.
 *  The format for division is noticeably non-standard.
 *  vdiv(op1, op2, quot, mod), where quot must be large enough
 *  to hold op1/op2 as expected, but mod must be as large as op1 + 2 words.
 *  The algorithm requires this much extra temp area to work in.
 * Of course, mod must also be able to hold the remainder.
 *
 *  Warning: a%b is defined inconsistently with some compilers,
 * for negative arguments.
 *
 *  no magnitude overflow checks are made.
 * In small memory model, malloc would fail first anyways.
 *
 *  Operands may be the same structure, as in vmul(x,x,y) for y = x^2.
 * The result can be the same structure as an operand for addition and
 * subtraction.  x -= y  vsub(x,y,x)     x = 2x  vadd(x,x,x)
 * The result must point to a different area for vmul and vdiv. */


#define VALUE_SIZE 2
#define VALUE_OFS 4
#define SIGN_OF(x) ((x)&=0x8000)
#define SIGN_CLR(x) ((x)&=0x7fff)
#define SIGN_SWAP(x) ((x)^=0x8000)

/* set magnitude, drop leading zeros */
static magset(v)
struct VINT *v;
{
short size = v->size;
int i;

SIGN_CLR(size);
for(i=size-1; i>=0; --i)
if(v->value[i]) break;
size = ++i;
SIGN_OF(v->size);
v->size |= size;
if(!size) v->size = 0;
} /* magset */

/* subtract from 0 (twos complement) */
static from0(v)
struct VINT *v;
{
int i;
unsigned short carry;
short size = v->size;

SIGN_CLR(size);
carry = 0;
for(i=0; i<size; ++i) {
unsigned short o1 = v->value[i];
unsigned short o3 = 0 - o1 - carry;
v->value[i] = o3;
carry = (o1|o3) >> 15;
}
} /* from0 */

/* magnitude (unsigned) add, called by vadd and vsub */
static uadd(v1, v2, v3, neg)
struct VINT *v1, *v2, *v3;
short neg; /* negate second argument */
{
short size1 = v1->size;
short size2 = v2->size ^ neg;
short usize1 = size1&0x7fff;
short usize2 = size2&0x7fff;
int i;
unsigned short carry;

/* make size2 the smaller */
if(usize1 < usize2) {
struct VINT *v = v1;
v1 = v2;
v2 = v;
i = size1; size1 = size2; size2 = i;
i = usize1; usize1 = usize2; usize2 = i;
}

v3->size = size1;

if(!((size1^size2)&0x8000)) {
/* same sign, ad magnitudes */
carry = 0;

for(i=0; i<usize2; ++i) {
unsigned short o1 = v1->value[i];
unsigned short o2 = v2->value[i];
unsigned short o3 = o1 + o2 + carry;
v3->value[i] = o3;
carry = (o1&o2 | ~o3&(o1^o2)) >> 15;
}

for(; i<usize1; ++i) {
unsigned short o1 = v1->value[i];
unsigned short o3 = o1 + carry;
v3->value[i] = o3;
carry = (~o3&o1) >> 15;
}

if(carry) { v3->value[i] = 1; ++ v3->size; }

} else {
/* different signs, subtract */
carry = 0;

for(i=0; i<usize2; ++i) {
unsigned short o1 = v1->value[i];
unsigned short o2 = v2->value[i];
unsigned short o3 = o1 - o2 - carry;
v3->value[i] = o3;
carry = (o3&~(o1^o2) | o2&~o1) >> 15;
}

for(; i<usize1; ++i) {
unsigned short o1 = v1->value[i];
unsigned short o3 = o1 - carry;
v3->value[i] = o3;
carry = (o3&~o1) >> 15;
}

if(carry) {
/* oops, second was really larger */
SIGN_SWAP(v3->size);
from0(v3);
} /* complement result */
} /* add or subtract */

magset(v3);
} /* uadd */

vadd(v1, v2, v3)
struct VINT *v1, *v2, *v3;
{
uadd(v1, v2, v3, 0);
} /* vadd */

vsub(v1, v2, v3)
struct VINT *v1, *v2, *v3;
{
uadd(v1, v2, v3, 0x8000);
} /* vsub */

vmul(v1, v2, v3)
struct VINT *v1, *v2, *v3;
{
short size1 = v1->size;
short size2 = v2->size;
short size3;
unsigned long l;
unsigned short carry;
int i, j;

v3->size = 0;
if(!size1 || !size2) return;
v3->size = (size1^size2) & 0x8000;
SIGN_CLR(size1);
SIGN_CLR(size2);
size3 = size1 + size2;
v3->size |= size3;

/* zero out result */
for(i=0; i<size3; ++i)
v3->value[i] = 0;

for(i=0; i<size1; ++i) {
carry = 0;
for(j=0; j<size2; ++j) {
l = (unsigned long)v1->value[i] * (unsigned long)v2->value[j] + carry + v3->value[i+j];
v3->value[i+j] = l;
carry = l>>16;
}
v3->value[i+j] = carry;
}

magset(v3);
} /* vmul */

vdiv(v1, v2, quot, rem)
struct VINT *v1, *v2, *quot, *rem;
{
short size1 = v1->size;
short size2 = v2->size;
unsigned short carry;
unsigned long l;
unsigned short dvsfwd; /* first word of divisor */
int i;

quot->size = rem->size = 0;
if(!size1 || !size2) return; /* division by 0 */
quot->size = (size1^size2) & 0x8000;
SIGN_CLR(size1);
SIGN_CLR(size2);

/* copy dividend into rem */
for(i=0; i<size1; ++i)
rem->value[i] = v1->value[i];
rem->size = size1; /* leav it unsigned for now */
dvsfwd = v2->value[size2-1];

if(size2 <= size1) { /* it might go in */

/* make division by small numbers relatively efficient */
if(size2 == 1) {
carry = 0;
for(i=size1-1; i>=0; --i) {
l = (((unsigned long)carry) << 16) |
v1->value[i];
carry = l % (unsigned long)dvsfwd;
quot->value[i] = l / (unsigned long)dvsfwd;
}

rem->value[0] = carry;
rem->size = 1;
quot->size |= size1;

} else {
/* long division */
short remlen, dvslen; /* lengths of remainder and divisor */
short divshift; /* shift divisor for long division routine */
unsigned short quotword;

/* Unlike the other routines, long division is byte oriented.
 * When dividing a long by a short, only the high byte can be trusted,
 * and even this might be off by 1.
 * To illustrate, think of bytes as decimal digits.
 * To extract the next digit in the quotient, divide the leading two
 * digits of the divisor (short) into the leading four digits of the remainder
 * (long).  At each step we assume the remainder is not 100 times the shifted
 * divisor.  Thus the 2-byte into 4-byte quotient has only the bottom word,
 * and we will take (and trust) the high order byte of this bottom word.
 * The actual divisor may be (nearly) as large as 1 + our two-digit
 * approximation.  This has the biggest effect when the divisor begins 10.
 * Say the divisor is really 10999....
 * The largest dividend is 100 times this divisor.
 * So our approximate quotient was 1099/10 = 109, yet it should have been 99.
 * This points out two things.  Worst case, our leading digit is off by 1.
 * Second, the approximate quotient may actually spill over into the third
 * digit (100).  We need to check for this.
 * Multiply, subtract, and note that the remainder is still below 100 times the
 * shifted divisor.
 * Since this C file emphasizes portability over efficiency,
 * we never read or write a short on an odd byte boundary.  This causes
 * some processors to explode.  We take the data one byte at a time. */

#define byteptr(v) ((unsigned char*)((v)->value))

/* the long division routine needs two extra 0 words leading the remainder */
rem->value[size1] = 0;
rem->value[size1+1] = 0;

dvslen = size2 * 2;
if((dvsfwd&0xff00) == 0) {
--dvslen;
dvsfwd <<= 8;
dvsfwd |= byteptr(v2)[dvslen-2];
}
remlen = size1*2;
if(byteptr(rem)[remlen-1] == 0) --remlen;

/* lengths now in byte resolution, is dividend still bigger than divisor? */
divshift = remlen - dvslen;
if(divshift >= 0) {
i = divshift/2;
quot->size |= i+1;
quot->value[i] = 0; /* in case we never set the top byte */

for(; divshift >= 0; --divshift, --remlen) {
unsigned short o1, o2, o3;
unsigned char *t;

t = byteptr(rem) + remlen - 3;
l = t[0] | ((long)t[1] << 8) | ((long)t[2] << 16) | ((long)t[3] << 24);
l /= dvsfwd;
quotword = (l>>8);
if(quotword == 0x100) --quotword;

/* multiply shifted divisor by quotword and subtract from remainder */
t = byteptr(rem) + divshift;
carry = 0;
for(i=0; i<size2; ++i, t += 2) {
o1 = ((unsigned short)t[1] << 8) | t[0];
l = (unsigned long)v2->value[i] * (unsigned long)quotword;
o2 = l;
/* two steps, subtrac carry and then o2 */
o3 = o1 - carry;
carry = (o3&~(o1^carry) | carry&~o1) >> 15;
o1 = o3;
o3 = o1 - o2;
carry += (o3&~(o1^o2) | o2&~o1) >> 15;
carry += (l>>16);
t[0] = o3, t[1] = o3>>8;
}
o1 = ((unsigned short)t[1] << 8) | t[0];
o2 = carry;
o3 = o1 - o2;
t[0] = o3, t[1] = o3>>8;
carry = (o3&~(o1^o2) | o2&~o1) >> 15;

if(carry) {
/* underflow, we guessed too high */
--quotword;

/* add the shifted divisor back in */
carry = 0;
t = byteptr(rem) + divshift;
for(i=0; i<size2; ++i, t += 2) {
o1 = ((unsigned short)t[1] << 8) | t[0];
o2 = v2->value[i];
o3 = o1 + o2 + carry;
t[0] = o3, t[1] = o3>>8;
carry = (o1&o2 | ~o3&(o1^o2)) >> 15;
}
o1 = ((unsigned short)t[1] << 8) | t[0];
o1 += carry;
t[0] = o1, t[1] = o1>>8;
} /* end underflow adjustment */

byteptr(quot)[divshift] = quotword;

if(byteptr(rem)[remlen+1]) {
/* internal error, that byte should be 0 */
quot->size = rem->size = 0;
return;
}
} /* end loop computing bytes of the quotient */
} /* dividend still bigger */
} /* long division */
} /* bigger dividend */

magset(quot);
magset(rem);

/* at this point the remainder is the remainder after dividing the unsigned numbers.
 * subtract from divisor if necessary.
 * Don't test quot->size negative, because the quotient may be 0,
 * the same as -0, and yet -1/256 should have a remainder of 255.
 * Note also that the divisor might be larger than the remainder. */
if(rem->size && (v1->size^v2->size)&0x8000) {
vsub(rem, v2, rem, v2->size&0x8000);
SIGN_SWAP(rem->size);
} /* complement remainder */
} /* vdiv */

